@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2011	PgPool Global Development Group
+ * Copyright (c) 2003-2012	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -797,14 +797,12 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 	len = ntohl(len);
 	len -= sizeof(len);
 
-	if (len <= 0)
+	if (len <= 0 || len >= MAX_STARTUP_PACKET_LENGTH)
 	{
 		pool_error("read_startup_packet: incorrect packet length (%d)", len);
-	}
-	else if (len >= MAX_STARTUP_PACKET_LENGTH)
-	{
-		pool_error("read_startup_packet: invalid startup packet");
 		pool_free_startup_packet(sp);
+		alarm(0);
+		pool_signal(SIGALRM, SIG_IGN);
 		return NULL;
 	}
 
@@ -1505,7 +1503,8 @@ void discard_persistent_db_connection(POOL_CONNECTION_POOL_SLOT *cp)
 }
 
 /*
- * do authentication for cp
+ * Do authentication. Assuming the only caller is
+ * *make_persistent_db_connection().
  */
 static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 {
@@ -1516,6 +1515,7 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 	char state;
 	char *p;
 	int pid, key;
+	bool keydata_done;
 
 	/*
 	 * read kind expecting 'R' packet (authentication response)
@@ -1647,8 +1647,12 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 	}
 
 	/*
-	 * read pid etc.
+	 * Read backend key data and wait until Reay for query arriving or
+	 * error happens.
 	 */
+
+	keydata_done = false;
+
 	for (;;)
 	{
 		status = pool_read(cp->con, &kind, sizeof(kind));
@@ -1661,6 +1665,7 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 		switch (kind)
 		{
 			case 'K':	/* backend key data */
+				keydata_done = true;
 				pool_debug("s_do_auth: backend key data received");
 
 				/* read message length */
@@ -1690,21 +1695,9 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 					return -1;
 				}
 				cp->key = key;
+				break;
 
-				/* read kind expecting 'Z' (ready for query) */
-				status = pool_read(cp->con, &kind, sizeof(kind));
-				if (status < 0)
-				{
-					pool_error("s_do_auth: error while reading kind");
-					return -1;
-				}
-
-				if (kind != 'Z')
-				{
-					pool_error("s_do_auth: expecting Z got %c", kind);
-					return -1;
-				}
-
+			case 'Z':	/* Ready for query */
 				/* read message length */
 				status = pool_read(cp->con, &length, sizeof(length));
 				if (status < 0)
@@ -1724,16 +1717,22 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 
 				pool_debug("s_do_auth: transaction state: %c", state);
 				cp->con->tstate = state;
+
+				if (!keydata_done)
+				{
+					pool_error("s_do_auth: ready for query arrived before receiving keydata");
+				}
 				return 0;
 				break;
 
 			case 'S':	/* parameter status */
-				pool_debug("s_do_auth: parameter status data received");
-
+			case 'N':	/* notice response */
+			case 'E':	/* error response */
+				/* Just throw away data */
 				status = pool_read(cp->con, &length, sizeof(length));
 				if (status < 0)
 				{
-					pool_error("s_do_auth: error while reading message length");
+					pool_error("s_do_auth: error while reading message length. kind:%c", kind);
 					return -1;
 				}
 
@@ -1746,7 +1745,7 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 				break;
 
 			default:
-				pool_error("s_do_auth: unknown response \"%c\" before processing BackendKeyData",
+				pool_error("s_do_auth: unknown response \"%c\" while processing BackendKeyData",
 						   kind);
 				break;
 		}
