@@ -814,22 +814,30 @@ static void stop_me(void)
 */
 static int read_pid_file(void)
 {
-	FILE *fd;
+	int fd;
+	int readlen;
 	char pidbuf[128];
 
-	fd = fopen(pool_config->pid_file_name, "r");
-	if (!fd)
+	fd = open(pool_config->pid_file_name, O_RDONLY);
+	if (fd == -1)
 	{
 		return -1;
 	}
-	if (fread(pidbuf, 1, sizeof(pidbuf), fd) <= 0)
+	if ((readlen = read(fd, pidbuf, sizeof(pidbuf))) == -1)
 	{
 		pool_error("could not read pid file as %s. reason: %s",
 				   pool_config->pid_file_name, strerror(errno));
-		fclose(fd);
+		close(fd);
 		return -1;
 	}
-	fclose(fd);
+	else if (readlen == 0)
+	{
+		pool_error("EOF detected while reading pid file as %s. reason: %s",
+				   pool_config->pid_file_name, strerror(errno));
+		close(fd);
+		return -1;
+	}
+	close(fd);
 	return(atoi(pidbuf));
 }
 
@@ -838,11 +846,11 @@ static int read_pid_file(void)
 */
 static void write_pid_file(void)
 {
-	FILE *fd;
+	int fd;
 	char pidbuf[128];
 
-	fd = fopen(pool_config->pid_file_name, "w");
-	if (!fd)
+	fd = open(pool_config->pid_file_name, O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR);
+	if (fd == -1)
 	{
 		pool_error("could not open pid file as %s. reason: %s",
 				   pool_config->pid_file_name, strerror(errno));
@@ -850,11 +858,26 @@ static void write_pid_file(void)
 		exit(1);
 	}
 	snprintf(pidbuf, sizeof(pidbuf), "%d", (int)getpid());
-	fwrite(pidbuf, strlen(pidbuf)+1, 1, fd);
-	if (fclose(fd))
+	if (write(fd, pidbuf, strlen(pidbuf)+1) == -1)
 	{
 		pool_error("could not write pid file as %s. reason: %s",
 				   pool_config->pid_file_name, strerror(errno));
+		close(fd);
+		pool_shmem_exit(1);
+		exit(1);
+	}
+	if (fsync(fd) == -1)
+	{
+		pool_error("could not fsync pid file as %s. reason: %s",
+				   pool_config->pid_file_name, strerror(errno));
+		close(fd);
+		pool_shmem_exit(1);
+		exit(1);
+	}	
+	if (close(fd) == -1)
+	{
+		pool_error("could not close pid file as %s. reason: %s",
+				   pool_config->pid_file_name, strerror(errno));
 		pool_shmem_exit(1);
 		exit(1);
 	}
